# coding=utf-8
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

import ai_integration.modes
from ai_integration.modes import *

DEBUG=False


# This is the only function that should be imported by the user's shim.
def _start_loop(inference_function=None, inputs_schema=None):
    loaded_module_names = ai_integration.modes.__all__

    if DEBUG:
        print('Loaded these integration module files:')
        print(loaded_module_names)

    all_modes_by_name = {}

    for module_name in loaded_module_names:
        this_module = getattr(ai_integration.modes, module_name)
        this_modules_modes = getattr(this_module, 'MODULE_MODES', None)
        if this_modules_modes is None:
            if DEBUG:
                print('module exports no modes: ' + module_name)
            continue
        for name in this_modules_modes:
            all_modes_by_name[name] = this_modules_modes[name]

    if DEBUG:
        print('All loaded modes by name:')
        print(all_modes_by_name)

    chosen_mode = os.environ.get('MODE')

    if chosen_mode is None:
        print('No mode chosen - defaulting to command line and listing other integration modes:')
        for mode_name in sorted(all_modes_by_name.keys()):
            hint = getattr(all_modes_by_name[mode_name], 'hint', None)
            if hint:
                print(' * MODE={}: {}'.format(mode_name, hint))
            else:
                print(' * MODE={}: (No hint attached to this function)'.format(mode_name))
        chosen_mode='command_line'

    if chosen_mode not in all_modes_by_name:
        print('MODE not found: '+str(chosen_mode))
        exit(1)

    mode_function = all_modes_by_name[chosen_mode]

    if inputs_schema is None:
        print('Warning: No inputs schema was provided to start_loop. Using empty inputs schema. Some integration modes may not work, if they use the inputs schema!')

    mode_function(inference_function=inference_function, inputs_schema=inputs_schema)
    print('Mode has exited cleanly: '+chosen_mode)
    exit(0)




# from https://stackoverflow.com/a/1057534/3064675

from os.path import dirname, basename, isfile
import glob
modules = glob.glob(dirname(__file__)+"/*.py")
__all__ = [ basename(f)[:-3] for f in modules if isfile(f) and not f.endswith('__init__.py')]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from PIL import Image
import io


def make_image_rgb_jpeg(img_bytes):
    im = Image.open(io.BytesIO(img_bytes))
    im_rgb = im.convert('RGB')

    with io.BytesIO() as output:
        # Use JPEG to maximize compatibility and tune it to not lose much quality
        im_rgb.save(output, format="JPEG", subsampling=0, quality=99)
        contents = output.getvalue()
        im_rgb.close()
        im.close()
        return contents



# coding=utf-8

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import io
import json
import pickle
import sys
import traceback


def check_inference_output(result):
    from PIL import Image
    if not isinstance(result, dict):
        print('Inference result is not a dict, total fail. Sad!')
        return

    if 'success' not in result:
        print('Error: "success" field is not passed.')

    if 'error' not in result:
        print('Error: "error" field is not passed.')

    if 'content-type' not in result:
        print('Error: "content-type" is not passed.')

    if 'data' not in result:
        print('Error: "data" is not passed.')

    if result['success'] == False:
        print('Model has indicated that it failed, success = False')
        print(result['error'])

    if result['success'] == True:
        print('Good, model has indicated that it succeeded!')

        if result['content-type'] == 'application/json':
            print('attempting to parse json response')
            parsed = json.loads(result['data'])
            print('parsed json successfully!')
            print(parsed)
        elif result['content-type'] == 'image/jpeg':
            print('attempting to parse jpeg response')
            image = Image.open(io.BytesIO(result['data']))
            print('opened image succesfully, I guess its ok')
        else:
            print('Unknown response content type:' + result['content-type'])


def test_inference_function(inference_function=None, inputs_schema=None):
    from PIL import Image

    solid_color_image = Image.new("RGB", (600, 400), (200, 200, 200))
    test_image_bytes = io.BytesIO()
    solid_color_image.save(test_image_bytes, format='JPEG')
    test_image_bytes = test_image_bytes.getvalue()

    print('Attempting inference a few times....')

    for i in range(5):
        print('Attempting inference...')
        try:
            result = inference_function({
                'image': test_image_bytes
            })
        except Exception:
            print('fail, inference produced an exception:')
            traceback.print_exc()
            sys.exit(1)
        print('Inference finished. Now inspecting the result.')
        check_inference_output(result)

    print('All done, quitting.')


def test_model_integration(inference_function=None, inputs_schema=None):
    print('Entering model integration test mode')

    test_inference_function(inference_function=inference_function, inputs_schema=inputs_schema)


test_model_integration.hint = 'Test running the function on a single gray image (docker run flag: -e MODE=test_model_integration )'


def test_single_image(inference_function=None, inputs_schema=None):
    print('Entering single input test mode')

    image_data = sys.stdin.read()
    print("the image data has type: " + str(type(image_data)))
    print("got %d bytes of image data" % len(image_data))
    print('Attempting inference...')
    try:
        result = inference_function({
            'image': image_data
        })
    except Exception:
        print('fail, inference produced an exception:')
        traceback.print_exc()
        sys.exit(1)
    print('Inference finished. Now inspecting the result.')
    check_inference_output(result)


test_single_image.hint = 'Single image: Pipe your image in with <, as in: nvidia-docker run --rm -i -e MODE=test_single_image ...reponame... < George-W-Bush.jpeg '


def test_inputs_dict_json(inference_function=None, inputs_schema=None):
    print('Entering test single inputs dict mode')

    json_data = sys.stdin.read()
    print("my json data: " + str(json_data))
    print("the json data has type: " + str(type(json_data)))
    print("got %d bytes of json data" % len(json_data))
    inputs_dict = json.loads(json_data)
    print("json after applying json.loads and stringified" + str(inputs_dict))
    print("the type of inputs dict is now: " + str(type(inputs_dict)))
    print('Attempting inference...')
    try:
        result = inference_function(inputs_dict)
    except Exception:
        print('fail, inference produced an exception:')
        traceback.print_exc()
        sys.exit(1)
    print('Inference finished. Now inspecting the result.')
    check_inference_output(result)


test_inputs_dict_json.hint = 'Pipe your dict in as JSON, as in: echo \'{"text":"i am some json data"}\' | nvidia-docker run --rm -i -e MODE=test_inputs_dict_json ...reponame...  '


def test_inputs_pickled_dict(inference_function=None, inputs_schema=None):
    print('Entering test pickled dict mode')

    pkl_data = sys.stdin.buffer.read()
    print("the pkl_data has type: " + str(type(pkl_data)))
    print("got %d bytes of pickle data" % len(pkl_data))
    inputs_dict = pickle.loads(pkl_data)
    print('Attempting inference...')
    try:
        result = inference_function(inputs_dict)
    except Exception:
        print('fail, inference produced an exception:')
        traceback.print_exc()
        sys.exit(1)
    print('Inference finished. Now inspecting the result.')
    check_inference_output(result)


test_inputs_pickled_dict.hint = 'Create a pickled dict of inputs, save as dict.pkl.  Then pipe your dict in, as in: cat dict.pkl | nvidia-docker run --rm -i -e MODE=test_inputs_pickled_dict ...reponame...  '

MODULE_MODES = {
    'test_model_integration': test_model_integration,
    'test_single_image': test_single_image,
    'test_inputs_dict_json': test_inputs_dict_json,
    'test_inputs_pickled_dict': test_inputs_pickled_dict
}



import sys
import time
import traceback
from contextlib import contextmanager
from threading import Thread
from multiprocessing import Queue

from .internal_methods import  _start_loop

global background_thread
background_thread = None
global inputs_queue
inputs_queue = None
global outputs_queue
outputs_queue = None

MAX_QUEUE_LEN = 1

global waiting_for_user_code_response
waiting_for_user_code_response = False


def background_thread_handler(inputs_queue, outputs_queue, input_schema):
    print('Hello from background thread')

    # Runs in background (library) thread.
    def fake_inference_function(inputs_dict):
        inputs_queue.put(inputs_dict)

        return outputs_queue.get(block=True)

    print('About to start loop in background thread...')
    _start_loop(inference_function=fake_inference_function, inputs_schema=input_schema)
    print('_start_loop in background thread exited, now quitting whole process.')
    sys.exit(0)


# Runs in main (user) thread.
def send_result(results_dict):
    global outputs_queue
    global waiting_for_user_code_response
    waiting_for_user_code_response = False
    outputs_queue.put(results_dict)


# Runs in main (user) thread.
@contextmanager
def get_next_input(*args, **kwds):
    global waiting_for_user_code_response
    global background_thread
    global outputs_queue
    global inputs_queue

    if background_thread is None:
        print('Launching background thread')
        # First time calling - need to launch background thread
        #print('args {}, kwd {} '.format(args, kwds))
        input_schema = kwds.get('inputs_schema')
        if input_schema is None:
            print('WARNING: no input_schema passed to context manager.')
        inputs_queue = Queue(MAX_QUEUE_LEN)
        outputs_queue = Queue(MAX_QUEUE_LEN)
        background_thread = Thread(target=background_thread_handler,
                                    args=(inputs_queue, outputs_queue, input_schema,))
        background_thread.daemon = True
        background_thread.start()
        print('Launched background thread')

    inputs_dict = inputs_queue.get(block=True)
    print('Main thread got an input, now handing it to the users code')
    try:
        waiting_for_user_code_response = True
        yield inputs_dict
        print('User code returned')


        # The user's code must now call send_result before this returns.

    except Exception as e:
        print('User code threw an exception')
        print("ai_integration Context handler caught error, sending it back over the queue...", e)
        result_data = {
            "content-type": 'text/plain',
            "data": None,
            "success": False,
            "error": traceback.format_exc()
        }
        send_result(result_data)

        time.sleep(3)  # Wait a bit to allow the background thread to do something with the output before quitting...

        raise e

    # User should have called send_result or crashed with an exception...
    if waiting_for_user_code_response:
        # If the context manager is re-entered without responding - the user messed up and we should log this
        message = "Caught Programming Error - no success or error was received before the context manager exited"
        print(message)
        result_data = {
            "content-type": 'text/plain',
            "data": None,
            "success": False,
            "error": message
        }
        send_result(result_data)
    else:
        print('Good, we have a valid response after the end of the context manager.')



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
# This module just re-exports start_loop to prevent importing a whole bunch of stuff if you do import * in the shim.

# noinspection PyUnresolvedReferences
from ai_integration.internal_methods import _start_loop as start_loop




import os
import traceback

from flask import Flask, request, render_template_string

app = Flask(__name__)
app.config['MAX_CONTENT_LENGTH'] = 100 * 1024 * 1024 # 100mb max size


global_inference_function = None
global_inputs_schema = None

@app.route('/', methods=['GET', 'POST'])
def hello():
    global global_inference_function
    global global_inputs_schema
    if request.method == 'POST':
        inputs_dict = {}
        for key, value in global_inputs_schema.items():
            if key in request.files:
                inputs_dict[key] = request.files[key].read()
            elif key in request.form:
                inputs_dict[key] = request.form[key]
            else:
                return 'Input not provided: '+key

        print('Attempting inference...')
        try:
            result = global_inference_function(inputs_dict)
        except Exception:
            return 'fail, inference produced an exception:'+traceback.print_exc()

        if result['success'] == True:
            return result['data'], 200, {'Content-Type': result.get('content-type')}
        else:
            return 'Model indicated that it failed: '+ repr(result)

    return render_template_string('''
    <!doctype html>
    <title>Model Test</title>
    <h1>Test Model</h1>
    <form method=post enctype=multipart/form-data target='result_iframe'>
       {% for key, value in input_schema.items() %}
       
          Input {{key}}: 
          {% if value.type == 'image' %}
            <input type=file name={{key}}>
          {% else %}
            <input type=text name={{key}}>
          {% endif %}
          
          
          <br/>
       {% endfor %}
      
      <input type=submit value=Upload>
    </form>
    <iframe name="result_iframe" src="" style="width:100%; height:600px;"></iframe>

    ''', input_schema=global_inputs_schema
    )

def http(inference_function=None, inputs_schema=None):
    global global_inference_function
    global global_inputs_schema
    global_inference_function = inference_function
    global_inputs_schema = inputs_schema
    os.environ['FLASK_ENV']='development'
    app.run(debug=True, host='0.0.0.0', use_reloader=False)


http.hint = 'Run a HTTP Server with a simple web UI, docker run flags: -e MODE=http -p 5000:5000'

MODULE_MODES = {
    'http': http
}



import setuptools
from setuptools import setup


# The text of the README file
README = open("./README.md", 'r').read()

# This call to setup() does all the work
setup(
    name="ai_integration",
    version="1.0.5",
    description="AI Model Integration for python",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/deepai-org/ai_integration",
    author="Deep AI, Inc",
    author_email="admin@deepai.org",
    license="Apache 2.0",
    classifiers=[
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=["Pillow", "Flask"],
    entry_points={
    },
)



# coding=utf-8

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import argparse
import sys
import traceback


def command_line(inference_function=None, inputs_schema=None):
    print('Entering command line mode')

    parser = argparse.ArgumentParser()

    for input in inputs_schema:
        parser.add_argument("--"+input, dest=input, type=argparse.FileType('rb'), required=True, help="Input file for {} input".format(input))

    parser.add_argument("--output", dest='output', type=argparse.FileType('wb'), required=False, help="Output file")

    args = parser.parse_args()

    inputs_dict = {}
    for input in inputs_schema:
        inputs_dict[input] = getattr(args, input).read()

    print('Attempting inference...')
    try:
        result = inference_function(inputs_dict)
    except Exception:
        print('fail, inference produced an exception:')
        traceback.print_exc()
        sys.exit(1)

    if result['success'] == True:
        print('Model indicated that it succeeded')
        if args.output:
            print('Writing output to file...')
            args.output.write(result['data'])
            args.output.close()
        else:
            print('No output file option passed, dumping to console')
            print(repr(result))
    else:
        print('Model indicated that it failed')
        print(repr(result))


command_line.hint = 'Pass file paths on the command line with options'

MODULE_MODES = {
    'command_line': command_line
}
